"""This module contains a class that can be used to store and retrieve measurements to/from a SQLite database."""

import sqlite3

from typing import Dict, Optional, Sequence, Union

class Database(object):
    """This class provides a wrapper around a SQLite database for storing/retrieving measurements."""

    def __init__(self, filename: str = "blueair.db") -> None:
        """Instantiate a new database using the provided filename."""
        self.db = sqlite3.connect(filename)
        self.db.row_factory = sqlite3.Row  # type: ignore
        self.cursor = self.db.cursor()

        self.cursor.execute("""
            CREATE TABLE IF NOT EXISTS measurements (
                id INTEGER PRIMARY KEY ASC,
                timestamp INTEGER NOT NULL UNIQUE,
                pm25 REAL NOT NULL,
                temperature REAL NOT NULL,
                humidity REAL NOT NULL,
                co2 REAL NOT NULL,
                voc REAL NOT NULL,
                all_pollution REAL NOT NULL
            );
        """)

        self.db.commit()

    def commit(self) -> None:
        """Commit changes to the database."""
        self.db.commit()

    def get_latest_timestamp(self) -> Optional[int]:
        """
        Retrieve the latest stored measurement timestamp from the database.

        Returns None if there is no such timestamp.
        """
        self.cursor.execute("""
            SELECT MAX(timestamp) FROM measurements;
        """)

        return self.cursor.fetchone()[0]  # type: ignore

    def get_all_measurements(self) -> Sequence[Dict[str, Union[str, int, float]]]:
        """Retrieve all stored measurements from the database."""
        self.cursor.execute("""
            SELECT * FROM measurements ORDER BY timestamp ASC;
        """)

        return (dict(row) for row in self.cursor.fetchall())  # type: ignore

    def insert_measurement(self, *, timestamp: int, pm25: float, temperature: float, humidity: float, co2: float, voc: float, all_pollution: float) -> None:
        """
        Insert a new measurement into the database.

        For performance reasons this does not actually commit the newly created
        record to disk. Call `commit` after inserting the last record of a
        batch to save them to disk.
        """
        self.cursor.execute("""
            INSERT INTO measurements (
                timestamp, pm25, temperature, humidity, co2, voc, all_pollution
            ) VALUES (
                ?, ?, ?, ?, ?, ?, ?
            );
        """, (timestamp, pm25, temperature, humidity, co2, voc, all_pollution))



"""This module contains the Blueair command line client."""

import argparse
import coloredlogs  # type: ignore
import logging
import matplotlib
import matplotlib.pyplot as pyplot
import pandas as pd
import tzlocal

from datetime import datetime, timedelta, timezone
from itertools import chain, groupby
from time import sleep
from typing import Any, Dict, Sequence, List

from .blueair import BlueAir
from .database import Database

logger = logging.getLogger(__name__)

def _tabularize(it: Sequence[Dict[str, Any]], titles: List[str], field_keys: List[str]) -> None:
    # Get maximum string lengths per dict key
    grouped = groupby(sorted(chain(*(i.items() for i in it)), key=lambda x: x[0]), lambda x: x[0])
    lengths = {k: max(len(str(w)) for _, w in v) for k, v in grouped}

    title_mapping = dict(zip(field_keys, titles))
    lengths = {k: max(v, len(title_mapping[k])) for k, v in lengths.items()}

    print("  ".join(title.ljust(lengths[field]) for title, field in zip(titles, field_keys)))
    print("  ".join("-" * lengths[field] for field in field_keys))

    for entry in it:
        print("  ".join(str(entry[field]).ljust(lengths[field]) for field in field_keys))

def _collect_measurements(blueair: BlueAir, database: Database, device_uuid: str) -> None:
    now = datetime.now(timezone.utc)
    start_timestamp = database.get_latest_timestamp() or int((now - timedelta(days=10)).timestamp())
    end_timestamp = int(now.timestamp())
    measurements = blueair.get_data_points_between(device_uuid, start_timestamp + 1, end_timestamp)

    for measurement in measurements:
        database.insert_measurement(**measurement)

    database.commit()

    logger.info(f"Collected {len(measurements)} new measurements")

def _plot_measurements(database: Database, filename: str) -> None:
    measurements = database.get_all_measurements()

    # get the local timezone
    zone = tzlocal.get_localzone().zone

    # Create a dataframe for plotting
    dataframe = pd.DataFrame(measurements)
    dataframe["timestamp"] = dataframe["timestamp"].apply(lambda timestamp: datetime.utcfromtimestamp(timestamp))
    dataframe["timestamp"] = dataframe["timestamp"].dt.tz_localize("UTC").dt.tz_convert(zone)  # type: ignore

    matplotlib.rcParams["timezone"] = zone  # type: ignore
    pyplot.style.use("bmh")  # type: ignore

    fig, axs = pyplot.subplots(4, 1, figsize=(10, 10), constrained_layout=True)  # type: ignore

    axs[0].plot(dataframe["timestamp"], dataframe["pm25"])
    axs[0].set_title("PM 2.5")
    axs[0].set_ylabel("ug/m3")
    axs[0].set_ylim(bottom=0)
    axs[0].set_facecolor("#ffffff")
    axs[0].margins(x=0.01, y=0.1)

    axs[0].axhspan(0, 10, facecolor="#8AE3CE40")
    axs[0].axhspan(10, 20, facecolor="#D1E5A840")
    axs[0].axhspan(20, 35, facecolor="#FEDC9A40")
    axs[0].axhspan(35, 80, facecolor="#FAC07E40")
    axs[0].axhspan(80, 1000, facecolor="#FA897E40")

    locator = matplotlib.dates.AutoDateLocator()  # type: ignore
    axs[0].get_xaxis().set_major_locator(locator)
    axs[0].get_xaxis().set_major_formatter(matplotlib.dates.ConciseDateFormatter(locator))  # type: ignore

    axs[1].plot(dataframe["timestamp"], dataframe["voc"])
    axs[1].sharex(axs[0])
    axs[1].set_title("VOC")
    axs[1].set_ylabel("ppb")
    axs[1].set_ylim(bottom=0)
    axs[1].set_facecolor("#ffffff")
    axs[1].margins(x=0.01, y=0.1)

    axs[1].axhspan(0, 200, facecolor="#8AE3CE40")
    axs[1].axhspan(200, 400, facecolor="#D1E5A840")
    axs[1].axhspan(400, 600, facecolor="#FEDC9A40")
    axs[1].axhspan(600, 800, facecolor="#FAC07E40")
    axs[1].axhspan(800, 10000, facecolor="#FA897E40")

    axs[2].plot(dataframe["timestamp"], dataframe["temperature"])
    axs[2].sharex(axs[0])
    axs[2].set_title("Temperature")
    axs[2].set_ylabel("C")
    axs[2].margins(x=0.01, y=0.1)

    axs[3].plot(dataframe["timestamp"], dataframe["humidity"])
    axs[3].sharex(axs[0])
    axs[3].set_title("Relative Humidity")
    axs[3].set_ylabel("%")
    axs[3].margins(x=0.01, y=0.1)

    pyplot.savefig(filename)
    pyplot.show()

def run() -> None:
    """Run the Blueair command line client."""
    # Create argument parser
    parser = argparse.ArgumentParser(description="An example application using the Python BlueAir client that collects and graphs measurements.")
    parser.add_argument("--email", help="The username for the BlueAir account")
    parser.add_argument("--password", help="The password for the BlueAir acount")
    parser.add_argument("--list-devices", action="store_true", help="List the available devices for the account and exit")
    parser.add_argument("--list-attributes", action="store_true", help="List the available attributes for the device and exit")
    parser.add_argument("--uuid", help="The device UUID to use for collecting measurements")
    parser.add_argument("--interval", type=int, metavar="N", help="Collect measurements every N seconds")
    parser.add_argument("--output", default="chart.png", help="The filename to use for the generated chart (defaults to chart.png)")
    parser.add_argument("--database", default="blueair.db", help="The filename to use for the SQLite database (defaults to blueair.db)")
    parser.add_argument("--verbose", action="store_true", help="Enable verbose logging")

    args = parser.parse_args()

    # Configure logger
    coloredlogs.install(
        level=args.verbose and logging.INFO or logging.WARNING,
        fmt="%(asctime)s %(name)s[%(process)d] %(levelname)s %(message)s"
    )

    if not args.email or not args.password:
        raise RuntimeError("Must provide both email and password")

    blueair = BlueAir(args.email, args.password)

    # Handle device list mode
    if args.list_devices:
        devices = blueair.get_devices()

        _tabularize(
            devices,
            ["UUID", "User ID", "MAC Address", "Device Name"],
            ["uuid", "userId", "mac", "name"]
        )

        exit()

    # Get UUID or fetch from device list if not specified
    uuid = args.uuid

    if not uuid:
        devices = blueair.get_devices()
        if not devices:
            raise RuntimeError("No devices found")
        elif len(devices) != 1:
            raise RuntimeError("Found multiple devices, use --uuid argument to specify which one to use")
        else:
            uuid = devices[0]["uuid"]

    # Handle attributes list mode
    if args.list_attributes:
        for key, value in blueair.get_attributes(uuid).items():
            print(f"{key}: {value}")

        exit()

    # Initialize database session
    database = Database(filename=args.database)

    # Loop if interval is specified
    if args.interval:
        while True:
            logger.info("Collecting measurements")
            _collect_measurements(blueair, database, uuid)

            logger.info("Generating chart")
            _plot_measurements(database, args.output)

            logger.info(f"Waiting {args.interval} seconds")
            sleep(args.interval)
    else:
        logger.info("Collecting measurements")
        _collect_measurements(blueair, database, uuid)

        logger.info("Generating chart")
        _plot_measurements(database, args.output)



"""This module provides the Blueair class to communicate with the Blueair API."""

import base64
import logging
import requests

from typing import Any, Dict, List, Mapping, Union
from typing_extensions import TypedDict

logger = logging.getLogger(__name__)

# The BlueAir API uses a fixed API key.
API_KEY = "eyJhbGciOiJIUzI1NiJ9.eyJncmFudGVlIjoiYmx1ZWFpciIsImlhdCI6MTQ1MzEyNTYzMiwidmFsaWRpdHkiOi0xLCJqdGkiOiJkNmY3OGE0Yi1iMWNkLTRkZDgtOTA2Yi1kN2JkNzM0MTQ2NzQiLCJwZXJtaXNzaW9ucyI6WyJhbGwiXSwicXVvdGEiOi0xLCJyYXRlTGltaXQiOi0xfQ.CJsfWVzFKKDDA6rWdh-hjVVVE9S3d6Hu9BzXG9htWFw"  # noqa: E501

MeasurementBundle = TypedDict("MeasurementBundle", {
    "sensors": List[str],
    "datapoints": List[List[Union[int, float]]]
})

MeasurementList = List[Mapping[str, Union[int, float]]]

def transform_data_points(data: MeasurementBundle) -> MeasurementList:
    """Transform a measurement list response from the Blueair API to a more pythonic data structure."""
    key_mapping = {
        "time": "timestamp",
        "pm": "pm25",
        "tmp": "temperature",
        "hum": "humidity",
        "co2": "co2",
        "voc": "voc",
        "allpollu": "all_pollution"
    }

    keys = [key_mapping[key] for key in data["sensors"]]

    return [dict(zip(keys, values)) for values in data["datapoints"]]

class BlueAir(object):
    """This class provides API calls to interact with the Blueair API."""

    def __init__(self, username: str, password: str, home_host: str = None, auth_token: str = None) -> None:
        """
        Instantiate a new Blueair client with the provided username and password.

        To optimize multiple instantiatons of this class a server hostname and
        authentication token can be provided. This will cause the client to
        reuse a session from a previously initialized client and saves up to
        two API calls.
        """
        self.username = username
        self.password = password
        self.home_host = home_host
        self.auth_token = auth_token

        if not self.home_host:
            self.home_host = self.get_home_host()

        if not self.auth_token:
            self.auth_token = self.get_auth_token()

    def get_home_host(self) -> str:
        """
        Retrieve the home host for the current username.

        The home host is the server that is used to interact with the Blueair
        device. It can be stored and reused to avoid requesting it again when
        reinitializing the class at a later time.
        """
        logger.info(f"GET https://api.blueair.io/v2/user/{self.username}/homehost/")

        response = requests.get(
            f"https://api.blueair.io/v2/user/{self.username}/homehost/",
            headers={
                "X-API-KEY-TOKEN": API_KEY
            }
        )

        return response.text.replace("\"", "")

    def get_auth_token(self) -> str:
        """
        Authenticate the user and retrieve the authentication token.

        The authentication token can be reused to prevent an additional network
        request when initializing the client.
        """
        logger.info(f"GET https://{self.home_host}/v2/user/{self.username}/login/")

        response = requests.get(
            f"https://{self.home_host}/v2/user/{self.username}/login/",
            headers={
                "X-API-KEY-TOKEN": API_KEY,
                "Authorization": "Basic " + base64.b64encode((self.username + ":" + self.password).encode()).decode()
            }
        )

        return response.headers["X-AUTH-TOKEN"]

    def api_call(self, path: str) -> Any:
        """
        Perform a Blueair API call.

        This is a low level function that is used by most of the client API calls.
        """
        logger.info(f"GET https://{self.home_host}/v2/{path}")

        return requests.get(
            f"https://{self.home_host}/v2/{path}",
            headers={
                "X-API-KEY-TOKEN": API_KEY,
                "X-AUTH-TOKEN": self.auth_token
            }
        ).json()

    def get_devices(self) -> List[Dict[str, Any]]:
        """
        Fetch a list of devices.

        Returns a list of dictionaries. Each dictionary will have a UUID key
        (the device identifier), a user ID, MAC address, and device name.

        Example response:

        [{"uuid":"1234567890ABCDEF","userId":12345,"mac":"1234567890AB","name":"My Blueair Device"}]
        """
        return self.api_call(f"owner/{self.username}/device/")

    # Note: refreshes every 5 minutes
    def get_attributes(self, device_uuid: str) -> Dict[str, Any]:
        """
        Fetch a list of attributes for the provided device ID.

        The return value is a dictionary containing key-value pairs for any
        available attributes.

        Note: the data for this API call is only updated once every 5 minutes.
        Calling it more often will return the same respone from the server and
        should be avoided to limit server load.
        """
        attributes = {}
        for item in self.api_call(f"device/{device_uuid}/attributes/"):
            attributes[item["name"]] = item["currentValue"]

        return attributes

    # Note: refreshes every 5 minutes, timestamps are in seconds
    def get_info(self, device_uuid: str) -> Dict[str, Any]:
        """
        Fetch device information for the provided device ID.

        The return value is a dictionary containing key-value pairs for the
        available device information.

        Note: the data for this API call is only updated once every 5 minutes.
        Calling it more often will return the same respone from the server and
        should be avoided to limit server load.
        """
        return self.api_call(f"device/{device_uuid}/info/")

    # Note: refreshes every 5 minutes
    def get_current_data_point(self, device_uuid: str) -> Mapping[str, Union[int, float]]:
        """
        Fetch the most recent data point for the provided device ID.

        Returns a dictionary containing a key-value mapping for the most recent
        measurements.

        Note: the data for this API call is only updated once every 5 minutes.
        Calling it more often will return the same respone from the server and
        should be avoided to limit server load.
        """
        data = self.api_call(f"device/{device_uuid}/datapoint/0/last/0/")

        results = transform_data_points(data)
        return results[-1]

    # Note: refreshes every 5 minutes
    def get_data_points_since(self, device_uuid: str, seconds_ago: int = 0, sample_period: int = 0) -> MeasurementList:
        """
        Fetch the list of data points between a relative timestamp (in seconds) and the current time.

        An optional sample period can be provided to group data points
        together. The minimum sample period size is 300 (5 minutes).

        Note: the data for the most recent data point is only updated once
        every 5 minutes.  Calling it more often will return the same respone
        from the server and should be avoided to limit server load.
        """
        data = self.api_call(f"device/{device_uuid}/datapoint/{seconds_ago}/last/{sample_period}/")

        results = transform_data_points(data)

        # Remove the last element because it does not have the final timestamp yet
        # TODO: Only remove if it is equal to end_timestamp
        results.pop()

        return results

    # Setting sample_period to a value higher than 300 (the minimum sample
    # period) will cause measurements to be averaged. Leave the sample period
    # to 0 to use the server's default period (300 seconds).  Calling this
    # function more than once per sample period will give the same results, so
    # make sure to throttle these calls to conserve API bandwidth.
    def get_data_points_between(self, device_uuid: str, start_timestamp: int, end_timestamp: int, sample_period: int = 0) -> MeasurementList:
        """
        Fetch the list of data points between two timestamps.

        The start and end timestamp are specified in seconds since the Unix
        epoch.

        An optional sample period can be provided to group data points
        together. The minimum sample period size is 300 (5 minutes).

        Note: the data for the most recent data point is only updated once
        every 5 minutes.  Calling it more often will return the same respone
        from the server and should be avoided to limit server load.
        """
        data = self.api_call(f"device/{device_uuid}/datapoint/{start_timestamp}/{end_timestamp}/{sample_period}/")

        results = transform_data_points(data)

        # Remove the last element because it does not have the final timestamp yet
        # TODO: Only remove if it is equal to end_timestamp
        results.pop()

        return results



#!/usr/bin/env python3

"""Package setup script."""

import codecs
import os
import re

from setuptools import find_packages, setup  # type: ignore
from typing import List

def get_absolute_path(*args: str) -> str:
    """Transform relative pathnames into absolute pathnames."""
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), *args)

def get_contents(*args: str) -> str:
    """Get the contents of a file relative to the source distribution directory."""
    return codecs.open(get_absolute_path(*args), "r", "UTF-8").read()

def get_version(*args: str) -> str:
    """Extract the version number from a Python module."""
    contents = get_contents(*args)
    metadata = dict(re.findall('__([a-z]+)__\\s+=\\s+[\'"]([^\'"]+)', contents))
    return metadata["version"]

def get_requirements(*args: str) -> List[str]:
    """Get requirements from pip requirement files."""
    requirements = set()

    with open(get_absolute_path(*args)) as handle:
        for line in handle:
            # Strip comments.
            line = re.sub(r"^#.*|\s#.*", "", line)

            # Ignore empty lines
            if line and not line.isspace():
                requirements.add(re.sub(r"\s+", "", line))

    return sorted(requirements)

setup(
    name="blueair",
    version=get_version("blueair", "__init__.py"),
    description="Unofficial Python client for BlueAir air purifiers",
    long_description=get_contents("README.rst"),
    url="https://blueair.readthedocs.io",
    author="Emil Loer",
    author_email="emil@codelle.com",
    license="MIT",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "blueair = blueair.cli:run",
        ]
    },
    package_data={
        "blueair": ["py.typed"]
    },
    install_requires=get_requirements("requirements.txt"),
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Communications",
        "Topic :: Internet",
        "Topic :: Office/Business",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: User Interfaces",
        "Topic :: System :: Hardware",
        "Topic :: System :: Logging",
        "Topic :: Utilities",
        "Typing :: Typed"
    ]
)
